# Copyright (C) 2022 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Any

import httpx
from cbra.ext import ioc

from ..apiclient import Consumer
from .credential import PicqerCredential
from .user import User


class PicqerClient(Consumer['PicqerClient']):
    __module__: str = 'molano.lib.picqer'
    domain: str
    http: httpx.AsyncClient

    def __init__(
        self,
        api_domain: str = ioc.environment('PICQER_API_DOMAIN'),
        credential: PicqerCredential = PicqerCredential.inject()
    ):
        self.credential = credential
        self.domain = api_domain

    def get_http_client_kwargs(self) -> dict[str, Any]:
        return {'base_url': f'https://{self.domain}/api/v1/'}

    # Implementation
    async def get_user(self, user_id: int) -> User:
        return await self.request(
            method='GET',
            response_model=User.parse_obj,
            url=f'users/{user_id}'
        )


# Copyright (C) 2022 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import httpx
from cbra.ext import ioc

from ..credential import ICredential


class PicqerCredential(ICredential):
    email: str
    key: str

    def __init__(
        self,
        api_email: str = ioc.environment('PICQER_API_EMAIL'),
        api_key: str = ioc.environment('PICQER_API_KEY')
    ):
        self.email = api_email
        self.key = api_key

    async def add_to_request(self, request: httpx.Request) -> None:
        request.headers.update({
            'User-Agent': f"Molano Internal Systems Gateway ({self.email})"
        })

    async def get_authentication(self) -> httpx.BasicAuth:
        return httpx.BasicAuth(username=self.key, password='X')


# Copyright (C) 2022 Cochise Ruhulessin
# 
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Any


GLOBAL_DEPENDENCIES: list[dict[str, Any]] = [
    {
        'kind': 'symbol',
        'spec': {
            'name': 'PicqerRepository',
            'qualname': 'molano.infra.PicqerRepository'
        }
    }
]


# pylint: skip-file
"""This test module exists so that at least one test is collected during
the unit test stage (the command errors when there are no tests that match
the include pattern).
"""
import unittest


class NoopUnitTestCase(unittest.TestCase):

    def test_noop(self):
        pass


if __name__ == '__main__':
    unittest.main()



# Copyright (C) 2022 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import os
from typing import Any


API_BASE_DOMAIN: str = os.environ['API_BASE_DOMAIN']

RESOURCE_SERVERS: dict[str, Any] = {
    'isg': {
        'server': f'https://isg.{API_BASE_DOMAIN}',
        'scope': {"molanoapis.com/internal"}
    },
}


# Copyright (C) 2022 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import datetime

import pydantic


class User(pydantic.BaseModel):
    iduser: int
    username: str
    firstname: str
    lastname: str
    emailaddress: str
    language: str | None
    admin: bool
    active: bool
    last_login_at: datetime.datetime | None
    created_at: datetime.datetime
    updated_at: datetime.datetime

    class Config:
        title: str = 'PicqerUser'


# Copyright (C) 2022 Cochise Ruhulessin
# 
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import logging
from typing import Any
from typing import NoReturn
from typing import TypeVar

import fastapi
import pydantic
from httpx import Response
from cbra.ext.service import CurrentService
from cbra.ext.service import ServiceClient


M = TypeVar("M", bound=pydantic.BaseModel)


class ServiceRepository:
    """A repository implementation that uses an internal service as
    its data source.
    """
    __module__: str = 'libwebid.lib'

    #: The service client instance that is used to create HTTP requests
    #: to internal services.
    client: ServiceClient

    #: The default logger instance. Is assumed to log messages to stdout.
    logger: logging.Logger = logging.getLogger('uvicorn')

    #: The logical name of the resource.
    resource: str

    @classmethod
    def inject(cls) -> Any:
        return fastapi.Depends(cls)

    def __init__(
        self,
        client: ServiceClient = CurrentService
    ) -> None:
        self.client = client

    async def retrieve(
        self,
        resource_path: str,
        response_model: type[M] = M,
        params: dict[str, str] | None = None
    ) -> M | None:
        """Lookup the given resource and parse a successful response
        with the specified `response_model`. Return ``None`` if the
        HTTP status code of the response is ``404``, otherwise attempt
        to parse the error artifact enclosed in the response body and
        raise the appropriate exception.
        """
        if not str.startswith(resource_path, '/'):
            resource_path = f'/{resource_path}'
        response = await self.client.get(
            resource=self.resource,
            path=resource_path,
            params=params
        )
        if response.status_code == 404:
            return None
        elif response.status_code >= 400:
            await self.on_request_exception(response)
            assert False # nosec
        try:
            return response_model.parse_obj(response.json()) # type: ignore
        except pydantic.ValidationError:
            self.logger.exception(
                "Unexpected response format (resource: %s, path: %s)",
                self.resource, resource_path
            )
            raise
        except Exception:
            self.logger.exception(
                "Caught fatal %s while parsing response (resoure: %s, path: %s)",
                self.resource, resource_path
            )
            raise

    async def on_request_exception(self, response: Response) -> NoReturn:
        response.raise_for_status()
        raise NotImplementedError



# Copyright (C) 2022 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import aorta
import pydantic

from ..purchaser import Purchaser
from ..supplierreference import SupplierReference
from ..warehousereference import WarehouseReference


class PurchaseOrderPurchased(aorta.Event):
    id: int = pydantic.Field(
        default=...,
        title="Purchase Order ID",
        description="The Picqer ID of the **Purchase Order**."
    )

    supplier: SupplierReference = pydantic.Field(
        default=...,
        title="Supplier",
        description=(
            "Describes the **Supplier** with which the **Purchase Order** "
            "was placed."
        )
    )

    supplier_document_id: str | None = pydantic.Field(
        default=None,
        title="Supplier Document ID",
        description=(
            "A document ID retained by the **Supplier** for this specific "
            "**Purchase Order**, such as an order or invoice number."
        )
    )

    warehouse: WarehouseReference = pydantic.Field(
        default=...,
        title="Warehouse",
        description=(
            "The **Warehouse** to which the goods included in the **Purchase "
            "Order** are expected to be shipped."
        )
    )

    purchaser: Purchaser = pydantic.Field(
        default=...,
        title="Purchaser",
        description=(
            "The person that placed the **Purchase Order** with the "
            "**Supplier**."
        )
    )


# Copyright (C) 2022 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import pydantic


class SupplierReference(pydantic.BaseModel):
    """A reference to a **Supplier**."""
    id: int | None = pydantic.Field(
        default=None,
        title="Supplier ID",
        description=(
            "The supplier identifier as retained in Picqer. "
            "This property may be `null` if the **Supplier** is "
            "not maintained in Picqer."
        )
    )

    name: str = pydantic.Field(
        default=...,
        title="Name",
        description="The current name of the **Supplier**."
    )


# Copyright (C) 2022 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Any

import fastapi
import httpx


class ICredential:
    """The base class for all HTTP credential implementations."""
    __module__: str = 'cbra.ext.api'

    @classmethod
    def inject(cls) -> Any:
        return fastapi.Depends(cls)

    async def add_to_request(self, request: httpx.Request) -> None:
        """Modify a :class:`httpx.Request` instance to include this
        credential.
        """
        pass

    async def get_authentication(self) -> httpx.Auth | None:
        return None

    async def preprocess_json(
        self,
        json: dict[str, Any] | list[Any]
    ) -> dict[str, Any] | list[Any]:
        """Preprocess the JSON payload of a request."""
        return json

    async def send(
        self,
        client: httpx.AsyncClient,
        request: httpx.Request
    ) -> httpx.Response:
        response = await client.send(request, auth=await self.get_authentication())
        return await self.process_response(client, response.request, response)

    async def process_response(
        self,
        client: httpx.AsyncClient,
        request: httpx.Request,
        response: httpx.Response
    ) -> httpx.Response:
        return response


# pylint: skip-file
"""This test module exists so that at least one test is collected during
the integration test stage (the command errors when there are no tests that
match the include pattern).
"""
import unittest


class NoopIntegrationTestCase(unittest.TestCase):

    def test_noop(self):
        pass


if __name__ == '__main__':
    unittest.main()



# Copyright (C) 2022 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Any
from typing import Callable
from typing import Generic
from typing import TypeVar

import fastapi
import httpx
from cbra.exceptions import NotFound

from .credential import ICredential
from .credential import NullCredential


C = TypeVar('C')
M = TypeVar('M')


class Consumer(Generic[C]):
    __module__: str = 'cbra.ext.api'
    credential: ICredential = NullCredential()

    @classmethod
    def inject(cls) -> C:
        return fastapi.Depends(cls)

    def get_http_client(self) -> httpx.AsyncClient:
        return httpx.AsyncClient(**self.get_http_client_kwargs())

    def get_http_client_kwargs(self) -> dict[str, Any]:
        raise NotImplementedError

    async def request(
        self,
        method: str,
        response_model: Callable[..., M],
        *args: Any,
        **kwargs: Any
    ) -> M:
        async with self.get_http_client() as client:
            request = await self.request_factory(
                client=client,
                method=method,
                *args, **kwargs
            )
            response = await self.credential.send(client, request)
        if response.status_code == 404:
            raise NotFound
        response.raise_for_status()
        return response_model(response.json())

    async def request_factory(
        self,
        client: httpx.AsyncClient,
        method: str,
        json: Any | None = None,
        *args: Any,
        **kwargs: Any
    ) -> httpx.Request:
        """Create a :class:`httpx.Request` instance."""
        if json is not None:
            kwargs['json'] = await self.preprocess_json(json)
        request = client.build_request(method=method, *args, **kwargs) # type: ignore
        await self.credential.add_to_request(request)
        return request

    async def preprocess_json(
        self,
        json: dict[str, Any] | list[Any]
    ) -> dict[str, Any] | list[Any]:
        """Preprocess the JSON payload of a request."""
        return await self.credential.preprocess_json(json)


# Copyright (C) 2022 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import pydantic


class WarehouseReference(pydantic.BaseModel):
    """A reference to a **Warehouse**."""
    id: int = pydantic.Field(
        default=...,
        title="Warehouse ID",
        description="The warehouse identifier as retained in Picqer. "
    )

    name: str = pydantic.Field(
        default=...,
        title="Name",
        description="The current name of the **Warehouse**."
    )


# Copyright (C) 2022 Cochise Ruhulessin
# 
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from molano.canon import PicqerUser
from molano.lib import ServiceRepository


class PicqerRepository(ServiceRepository):
    resource: str = 'isg'

    async def get_user(self, user_id: int) -> PicqerUser | None:
        return await self.retrieve(
            resource_path=f'users/{user_id}',
            response_model=PicqerUser
        )